#!/usr/bin/python

__rcs_id__ = """$Id: install_routines.py 935736 2013-10-25 03:05:02Z aprisyazhnyuk $"""

from xml.dom import minidom as dom
import tarfile
import socket
import os
import sys
import re
import urllib
import stat
import shutil
import errno

from poaupdater import uPEM, uSysDB, uLogging, uDBSchema, uPDLDBSchema, uDLModel, uUtil, uFakePackaging, uPackaging, uBuild, uFSReader, uConfig, openapi, simpleInstallerUI, PEMVersion, uAPS, uDialog

if 'win32' == sys.platform:
	MODULE_DIR = os.path.realpath(os.path.dirname(sys.executable))
else:
	MODULE_DIR = os.path.realpath(os.path.dirname(__file__))

__pychecker__ = "--blacklist=['sets'] --unusednames=__rcs_id__,dummy"
any_any_any = ('any', 'any', 'any')

platforms = {
  any_any_any : [ ('i386', 'any', 'any') ],
  ('i386', 'any', 'any') : [  ('i386', 'Linux', 'any'),  ('i386', 'Win32', 'any'), ('i686', 'any', 'any') ],

  ('i686', 'any', 'any') : [ ('i686', 'Linux', 'any') ],
  ('i686', 'Linux', 'any'): [ ('i686', 'RHEL', '6')],

  ('i386', 'Linux', 'any') : [ ('i386', 'RHES', '4'), ('i386', 'RHEL', '5')],

 ('i386', 'RHES', '4') : [ ('x86_64', 'RHES', '4') ],
 ('i386', 'RHEL', '5') : [ ('x86_64', 'RHEL', '5') ],
 ('i686', 'RHEL', '6') : [ ('x86_64', 'RHEL', '6') ],
 ('i386', 'Win32', 'any') : [ ('i386', 'Win32', '5.x') ],
 ('i386', 'Win32', '5.x') : [ ('i386', 'Win32', '5.0') ],
 ('i386', 'Win32', '5.0') : [ ('i386', 'Win32', '5.1') ],
 ('i386', 'Win32', '5.1') : [ ('i386', 'Win32', '5.2') ],
 ('i386', 'Win32', '5.2') : [ ('i386', 'Win32', '6.0'), ('x86_64', 'Win32', '5.2') ],
 ('i386', 'Win32', '6.0') : [ ('x86_64', 'Win32', '6.0') ],
 ('x86_64', 'Win32', '6.0') : [ ('x86_64', 'Win32', '6.1') ],
 ('x86_64', 'Win32', '6.1') : [ ('x86_64', 'Win32', '6.2') ],
}

base_pkgs = [
   ("other", "SoLoader"),
   ("other", "libplatform"),
   ("sc", "chief"),
   ("sc", "Kernel"),
   ("sc", "ppm"),
   ("other", "pleskd"),
   ("other", "pem_config")
]

pgp_key_id = "00d214045ef44d2e"

OID_ADMIN_ACCOUNT = 1
OID_ADMIN_USER = 1

OID_SYSTEM_ACCOUNT = 2
OID_SC_USER = 2


predefined_roles = [
	( 1, 'Account Administrator', 'This role defines privileges for provider', 1),
	( 2, 'Account Administrator', 'This role defines privileges for resellers', 1),
	( 3, 'Account Administrator', 'This role defines privileges for clients', 1),

	( 1, 'Pleskd role.', 'This defines privileges of pleskd system user ', 1),

	( 1, 'Staff member defaults', 'This role defines default privileges for staff members', 2)
]

user_repo_types = [
	( "pgpkey", 'PGP Key'),
	( "ca_certificate", 'CA certificate'),
	( "certificate", 'Certificates'),
	( "private_key", 'Private keys')
]

def __findPlatform(parent, tofind):
	if not platforms.has_key(parent):
		return None
	for p in platforms[parent]:
		if p == tofind:
			return [p]

	for p in platforms[parent]:
		rv = __findPlatform(p, tofind)
		if rv is not None:
			return [p] + rv
	return None

def getPlatformLine(platform):
	ptuple = platform.arch, platform.os, platform.osver
	return [ any_any_any ] + __findPlatform(any_any_any, ptuple)

class Installation:
	def __init__(self, platform, build, props, progress):
		if props.get("backup-helper"):
			self.migrate = True
		else:
			self.migrate = False

		self.progress = progress
		self.platform = platform
		self.build = build
		if not self.migrate:
			self.buildinfo = uDLModel.getCumulativeUpdateInstructions([build])
		self.rootpath = props.get("rootpath", "/usr/local/pem")
		self.platform_line = getPlatformLine(platform)

		self.admin_login = props.get("username", 'admin')
		self.admin_passwd = props.get("password", '1q2w3e')
		self.hostname = props.get("hostname", socket.gethostname())

		self.sysuid = props.get('sysuserid')
		self.sysgid = props.get('sysgroupid')

		first_name = props.get("first_name", "Charlie")
		last_name = props.get("last_name", "Root")

		email = props.get("email", 'root@' + self.hostname)

		self.provider_contacts = first_name, last_name, email

		self.importedBasePackages = {}

		self.properties = props.copy()
		self.latestPackages = None

		self.system_login = props.get("system_login", "system_for_scs")
		self.system_password = props.get("system_password", uUtil.generate_random_password(32))

def initialize(platform, installation_props, build, iprogress, config):
	from poaupdater import uSysDB
	if not hasattr(config, 'database'):
		config.database = uConfig.Config()
	config.database.name = installation_props['db']
	config.database.user = installation_props['dsn_login']
	config.database.password = installation_props['dsn_passwd']
	uLogging.init(config)
	uSysDB.init(config)

	# need to clean connection pool
	uSysDB.disconnect()

	return Installation(platform, build, installation_props, iprogress)

def winConfig():
	return uConfig.Config(uConfig.getCommonConfigArgs() + [("repo-base-url", str, None, "")], [ "--verbose" ])

def initialize_from_props(props, iprogress):
	platform = uBuild.Platform(props["opsys"], props["osver"], props["arch"])
	source_root = props["source_dir"]
	rootpath = props['rootpath']
	if iprogress is None:
		iprogress = simpleInstallerUI.InstallationProgress()
	build = uFSReader.getBuildFromFS(source_root, source_root, calculateChecksums = False)
	install_dir = os.path.join(rootpath, 'install')
	uUtil.saveObj(build, install_dir, 'build.dat')
	uUtil.saveObj(props, install_dir, 'properties.dat')

	return initialize(platform, props, build, iprogress, winConfig())

def initialize_from_FS(rootpath, iprogress):
	if iprogress is None:
		iprogress = simpleInstallerUI.InstallationProgress()
	install_dir = os.path.join(rootpath, 'install')
	build = uUtil.loadObj(install_dir, 'build.dat', None)
	props = uUtil.loadObj(install_dir, 'properties.dat', None)
	platform = uBuild.Platform(props["opsys"], props["osver"], props["arch"])
	return initialize(platform, props, build, iprogress, winConfig())

def createContact(con, first_name, last_name, email):
	display_name = '%s %s' % (first_name, last_name)
	cur = con.cursor()
	cur.execute("INSERT INTO contacts (display_name) VALUES (%s)", display_name)
	contact_id = uSysDB.get_last_inserted_value(con, "contacts")
	cur.execute("INSERT INTO contact_properties(contact_id, name, value) VALUES (%s, %s, %s)", contact_id, "first_name", first_name)
	cur.execute("INSERT INTO contact_properties(contact_id, name, value) VALUES (%s, %s, %s)", contact_id, "last_name", last_name)
	cur.execute("INSERT INTO contact_properties(contact_id, name, value) VALUES (%s, %s, %s)", contact_id, "email", email)

	return contact_id

def createAddress(con):
	cur = con.cursor()
	cur.execute("INSERT INTO addresses DEFAULT VALUES")
	return uSysDB.get_last_inserted_value(con, "addresses")

def createRootAccount(con, account_id, atype, first_name, last_name, email):
	adm_contact_id = createContact(con, first_name, last_name, email)
	bill_contact_id = createContact(con, first_name, last_name, email)
	tech_contact_id = createContact(con, first_name, last_name, email)
	address_id = createAddress(con)
	cur = con.cursor()
	accounts_serial = uDBSchema.getTable("accounts_serial", con)
	accounts_serial.identityInsertBegin(con)
	con.insertRecord("accounts_serial", account_id = account_id)
	cur.execute("INSERT INTO accounts (account_id, owner_id, account_type, address_id, adm_contact_id, bill_contact_id, tech_contact_id, path, rt_instance_id) VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s)", account_id, account_id, atype, address_id, adm_contact_id, bill_contact_id, tech_contact_id, 'x' + str(account_id), 0)
	accounts_serial.identityInsertEnd(con)
	cur.execute("DELETE FROM accounts_serial WHERE account_id = %s", account_id)


def createRootAccountUser(con, account_id, user_id, first_name, last_name, email, login, password):
	"""
	Creates a main system user
	:param con: a database connection
	:param account_id: id of a user account
	:param user_id: id of a user
	:param first_name: user's first name
	:param last_name: user's last name
	:param email: user's e-mail
	:param login: user's login
	:param password: unencrypted user's password
	"""
	
	cur = con.cursor()
	
	contact_id = createContact(con, first_name, last_name, email)	
	cur.execute("INSERT INTO identities (system_id, login, login_lowered) VALUES (NULL, %s, lower(%s))", login, login)
	identity_id = uSysDB.get_last_inserted_value(con, "identities")
	cur.execute("INSERT INTO local_identities (identity_id, pwd_type, pwd_hash) VALUES (%s, NULL, %s)", identity_id, password)
	users = uDBSchema.getTable("users", con)
	users.identityInsertBegin(con)
	cur.execute("INSERT INTO users(user_id, scope_id, member_id, auth_identity_id, contact_id, type) VALUES (%s, %s, %s, %s, %s, 'member')", user_id, account_id, user_id, identity_id, contact_id)
	users.identityInsertEnd(con)
	
	#member_id's should start from root account user_id
	members_serial = uDBSchema.getTable("members_serial", con)
	members_serial.identityInsertBegin(con)
	con.insertRecord("members_serial", member_id = user_id)
	members_serial.identityInsertEnd(con)
	cur.execute("DELETE FROM members_serial WHERE member_id = %s", user_id)
	
	#--- Added to detect id mixup. Must be removed this before release.	
	if uSysDB.DBType == uSysDB.PgSQL:
		cur.execute("SELECT setval('users_seq', 100)")
		cur.execute("SELECT setval('members_serial_seq', 200)")		
	#---


def createPredefinedRoles(con):
	cur = con.cursor()
	for role_type, name, description, admin in predefined_roles:
		cur.execute("INSERT INTO roles (role_type, name, description, account_id, admin) VALUES (%s, %s, %s, %s, %s)", role_type, name, description, OID_ADMIN_ACCOUNT, admin)

def createFirstUsers(con, installation):
	first_name, last_name, email = installation.provider_contacts
	createRootAccount(con, OID_ADMIN_ACCOUNT, 1, first_name, last_name, email)
	createRootAccountUser(con, OID_ADMIN_ACCOUNT, OID_ADMIN_USER, first_name, last_name, email, installation.admin_login, installation.admin_passwd)
	createRootAccount(con, OID_SYSTEM_ACCOUNT, 4, "System", "User", email)
	createRootAccountUser(con, OID_SYSTEM_ACCOUNT, OID_SC_USER, "System", "User", email, installation.system_login, installation.system_password)

	createPredefinedRoles(con)
	cur = con.cursor()
	for user_id in (OID_ADMIN_USER, OID_SC_USER):
		cur.execute("INSERT INTO role_users(role_id, user_id) SELECT role_id, %s FROM roles WHERE account_id = %s AND role_type = 1 AND admin = 1", user_id, OID_ADMIN_ACCOUNT)

	cur.execute("INSERT INTO interface_properties(user_id, name, value, type) VALUES (%s, 'language', 'English', 1)", OID_ADMIN_USER)

def completeFirstUsersCreation(installation):
	"""
	This function finishes creating of system users by hashing their passwords
	:param installation: installation object
	"""
	api = openapi.OpenAPI()
	api.pem.setMemberPassword(login = installation.admin_login, password = installation.admin_passwd)

def findSuitablePackage(platform_line, packages):
	for pkg in packages:
		pp = pkg.package.platform
		if (pp.arch, pp.os, pp.osver) in platform_line:
			return pkg
	return None

def findBasePackages(installation):
	platform = installation.platform
	packages = {}

	for pkg in base_pkgs:
		packages[pkg] = []

	contents = installation.build.contents
	for p in contents:
		if (p.ctype, p.name) in packages:
			packages[(p.ctype, p.name)].append(contents[p])

	rv = []
	for pkg in packages:
		p = findSuitablePackage(installation.platform_line, packages[pkg])

		if p is None:
			raise Exception("Cannot find %s-%s, suitable for %s-%s-%s" % (pkg[0], pkg[1], platform.arch, platform.os, platform.osver))
		rv.append(p)

	return rv

def reporthook(count, blockSize, totalSize): 
	if totalSize == 0:
			return

def retrieveFile(repo_url, file, install_dir):
	url = repo_url + "/" + file
	target = os.path.join(install_dir, file)
	if os.path.exists(target):
		return target
	else:
		uLogging.debug("Retrieving " + url)

	urllib._urlopener = urllib.URLopener()
	download = urllib.urlretrieve(url, reporthook=reporthook)

	dir = os.path.dirname(target)
	if not os.path.exists(dir):
		os.makedirs(dir)
	shutil.copy2(download[0], target)
	os.chmod(target, stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH)		

	return target

def downloadDistribution(progress, config):
	repo_url = config.repobaseurl
	local_install = repo_url.startswith("file://")
	install_dir = MODULE_DIR

	if local_install:
		uLogging.notice("Installing from local directory %s" % repo_url)
		return
	else:
		uLogging.notice("Downloading distribution from repo %s" % repo_url)	

	progress.new_phase("Downloading distribution index")

	import simplejson as json
	# retrieve indexed packages from repository
	index_files = [
	# 	index		rpm-platform
	("rindex-RHEL-5.json", ("RHEL", "5")),
	("rindex-RHEL-6.json", ("RHEL", "6")),
	("rindex-RHES-4.json", ("RHES", "4")),
	("rindex-Win32-5.2.json", None),
	]

	# collect whole package index for all platforms
	package_index = {}

	total = len(index_files)
	for i in range(total):
		index_file, rpm_platform = index_files[i]
		try:
			progress.set_progress(100 * i / total, "%s" % index_file)

			# download index
			file = retrieveFile(repo_url, index_file, install_dir)
		except Exception, e:
			uLogging.warn("Failed to retrieve %s: %s" % (index_file, e))
			continue

		f = open(file)
		data = f.read()
		obj = json.loads(data)
		f.close()

		for module, packages in obj.iteritems():
			if module in package_index:
				package_index[module] += packages
			else:
				package_index[module] = packages

	# download all the files mentioned in index
	modules = [p for p in package_index.keys() if "UNDEFINED" != p]
	modules += ["Platform"]

	packages = []
	for module in package_index:
		packages += package_index[module]

	package_files = [p.split(" ")[0] for p in packages]

	# TODO create better index and remove this
	package_files = [p.replace("\\", "/") for p in package_files]
	package_files = [p.replace("//", "/") for p in package_files]

	mdls = [m + ".mdl" for m in modules]
	files_to_download = package_files + mdls

	progress.new_phase("Downloading distribution data")

	total = len(files_to_download)
	for i in range(total):
		to_download = files_to_download[i]
		progress.set_progress(100 * i / total, "%s" % to_download)
		downloaded = retrieveFile(repo_url, to_download, install_dir)

		# TODO verify checksum of downloaded file

def deployBaseFiles(installation):
	uLogging.debug("Deploying base files")
	to_install = findBasePackages(installation)

	total = len(to_install)
	num = 0
	for p in to_install:
		tarball = p.findTarball(installation.build)
		if tarball is None:
			num += 1
			continue
		installation.progress.set_progress(100 * num / total, "Unpacking %s" % p.package)
		arc = tarfile.open(tarball)
		p.debug = 1
		p.errorlevel = 2
		for tarinfo in arc:
			arc.extract(tarinfo, path = installation.rootpath)
		num += 1

	uPackaging.update_stored_installer([installation.build], installation.rootpath)

def importPGPKey(props, con):
	f = open(os.path.join(props['source_dir'], 'plesk2@plesk.com.pub'), 'r')
	key = f.read()
	f.close()
	cur = con.cursor()
	cur.execute("INSERT INTO user_repo(user_id, type, name, description, item_size) VALUES (1, 'pgpkey', %s, 'PA PGP key', 0)", pgp_key_id)
	r_id = uSysDB.get_last_inserted_value(con, "user_repo")
	cur.execute("INSERT INTO user_repo_data(item_id, data) VALUES (%s, %s)", r_id, uSysDB.toRaw(key))

	con.commit()

class DBObjsCounter(uUtil.CounterCallback):
	def __init__(self, progress):
		uUtil.CounterCallback.__init__(self)
		self.progress = progress

	def new_item(self, item):
		t, name = item
		if t == 'dml':
			msg = "Inserting data"
		else:
			msg = "Creating %s %s" % (t, name)
		self.progress.set_progress(self.percents(), msg)

def createDBStructure(installation):
	uLogging.debug("Creating database structure")
	chief_packages = []
	for p in installation.build.contents:
		if p.name == 'chief' and p.ctype == 'sc':
			chief_packages.append(installation.build.contents[p])

	chief = findSuitablePackage(installation.platform_line, chief_packages)
	uPDLDBSchema.createDBStructureByManifest(os.path.join(installation.build.topdir, chief.manifest_file), DBObjsCounter(installation.progress))

def importBaseManifests(installation):
	base_packages = findBasePackages(installation)

	platform_cache = {}
	for pkg in base_packages:
		installation.importedBasePackages[(pkg.package.name, pkg.package.ctype)] = uFakePackaging.importPackage(os.path.join(installation.build.topdir, pkg.manifest_file), platform_cache)

def __registerSubPlatforms(con, parent):
	parent_platform = uBuild.Platform(parent[1], parent[2], parent[0])
	for p in platforms[parent]:
		platform = uBuild.Platform(p[1], p[2], p[0])
		uPackaging.registerPlatform(platform, parent_platform, con)
		if platforms.has_key(p):
			__registerSubPlatforms(con, p)

def registerPlatforms(con):
	# any-any-any:
	pforms = uDBSchema.getTable("platforms", con)
	pforms.identityInsertBegin(con)
	cur = con.cursor()
	cur.execute("INSERT INTO platforms(platform_id, arch, opsys, osrel, parent_id, l, r) VALUES(1, 'any', 'any', 'any', 1, 1, 2)")
	pforms.identityInsertEnd(con)

	__registerSubPlatforms(con, any_any_any)

def registerRepoTypes(con):
	cur = con.cursor()
	for repo_type, descr in user_repo_types:
		cur.execute("INSERT INTO user_repo_types(type, description) VALUES (%s, %s)", repo_type, descr)


def initResourcesTables(con):
	subscriptions_serial = uDBSchema.getTable("subscriptions_serial", con)
	subscriptions_serial.identityInsertBegin(con)
	con.insertRecord("subscriptions_serial", sub_id = 1)
	subscriptions_serial.identityInsertEnd(con)
	cur = con.cursor()
	cur.execute("DELETE FROM subscriptions_serial WHERE sub_id = %s", 1)
	con.insertRecord("subscriptions", sub_id = 1, owner_id = OID_ADMIN_ACCOUNT, name="License key", description = "Parallels Automation License Key")

	resource_types = uDBSchema.getTable("resource_types", con)
	resource_types.identityInsertBegin(con)
	con.insertRecord("resource_types", rt_id = -1, restype_name = 'Fake RT parent', parent_id = -1)
	resource_types.identityInsertEnd(con)

def registerAtomicActions(installation):
	ids = []
	for udl in installation.build.instructions:
		fn = os.path.join(installation.build.topdir, udl)
		doc = dom.parse(fn)
		for tagname in ("EXEC", "SQL", "INSTRUCTION"):
			ids += [ node.getAttribute("id") for node in doc.getElementsByTagName(tagname) if node.getAttribute("id") ]

	if ids:
		con = uSysDB.connect()
		for aid in ids:
			con.insertRecord("updater_atomic_actions", action_id = aid)
		con.commit()

def restoreOriginalDB(installation_props):
	bu_helper = installation_props['backup-helper']
	host = installation_props['hostname']

	dump_file = bu_helper.getDBDump()
	out, err, status = uUtil.readCmdExt(["pg_restore", "-U", "postgres", "-h", host, "-d", "plesk", "--format=c", dump_file])
	# here we ignore error code 1, because some of the DB object cannot be recreated, but there is no easy way to ignore only them
	if status and 1 != status:
		raise Exception("Failed to restore original DB: %s %s %s" % (out, err, status))


def checkOriginalConfig(installation):
	host_id = 1

	con = uSysDB.connect()
	cur = con.cursor()
	cur.execute("SELECT component_id, sc_id, name FROM services WHERE host_id = %s AND name NOT IN ('PrivacyProxy')", host_id)

	customer_services = [(row[0], row[1]) for row in cur.fetchall()]
	if len(customer_services) > 0:
		raise Exception("Original MN has customer services installed. Migration is not possible. Please remove services manually and recreate backup for migration.")

	con.commit()
	con.close()


def restoreOriginalConfig(installation):
	"""
	Sets some POA file owner and access permissions to suite latest POA requirements
	:param installation: installation configuration structure
	:raise: standard python Exception
	"""
	con = uSysDB.connect()
	cur = con.cursor()
	cur.execute('DELETE FROM openapi_requests WHERE request_id IN (\'create_repos\', \'encrypt_password\')')  # workaround for POA-75791
	con.commit()

	etc_objects = ['Kernel.conf', 'openapi', 'pleskd.props']

	for tgt in etc_objects:
		tpath = os.path.join(installation.rootpath, 'etc', tgt)
		if os.path.exists(tpath):
			os.chown(tpath, installation.sysuid, installation.sysgid)
		else:
			raise Exception("Migration failed due to essential path does not exist: %s" % tpath)

	for tgt in ['odbc.ini', 'odbcinst.ini']:
		os.chmod(os.path.join(installation.rootpath, 'etc', tgt), 0644)


def changePlatform(installation):
	con = uSysDB.connect()

	old_platform_id = uPackaging.getHostPlatform(con, 1)
	new_platform_id = uPackaging.getPlatformId(con, (installation.platform.arch, installation.platform.os, installation.platform.osver))
	platform_desc = "%s %s %s" % (installation.platform.arch, installation.platform.os, installation.platform.osver)

	uPackaging.updatePackagesRecords(con, 1, old_platform_id, new_platform_id)

	cur = con.cursor()
	cur.execute("UPDATE hosts SET platform_id = %s, platform_desc = %s WHERE host_id = %s", new_platform_id, platform_desc, 1)
	con.commit()
	con.close()


def restoreMN(installation):
	installation.progress.set_progress(0, "Starting the minimal system")
	uPEM.startMN(True)

	uPEM.wait_for_scs(["chief", "ppm"], monitor = SCWaiter(["chief", "ppm"], installation.progress, 0, 2))

	installation.progress.set_progress(10, "Reinstalling packages")
	mn_host_id = 1

	all_packages = uPackaging.listInstalledPackagesOnHost(mn_host_id)
	for pkg in all_packages:
		try:
			uPackaging.reinstallPackageToHost(mn_host_id, pkg.name, pkg.ctype)
		except uPackaging.ComponentNotFound, e:
			uLogging.warn("Package '%s' re-installation ignored due to: %s", pkg.name, e)
		except uUtil.ExecFailed, e:
			uLogging.warn("Package '%s' failed to re-install due to error: %s", pkg.name, e)
			if uDialog.askYesNo("Ignore this package and continue migration?"):
				#TODO: We have to clean up restored POA database somehow like 'ppm_ctl remove' does
				pass
			else:
				raise

	uPEM.stopMN(True)

	installation.progress.set_progress(90, "Starting management node")
	uPEM.startMN()

	uPEM.startUI(mn_host_id)
	installation.progress.set_progress(100)

def initDB(installation):
	uLogging.debug("initializing DB")
	con = uSysDB.connect()

	progress = installation.progress
	progress.set_progress(0, "Creating system users")
	createFirstUsers(con, installation)

	progress.set_progress(10, "Importing PGP key for packaging")
	registerRepoTypes(con)
	importPGPKey(installation.properties, con)

	progress.set_progress(20, "Registering known platforms")
	registerPlatforms(con)
	progress.set_progress(30, "Creating system resource types")
	initResourcesTables(con)

	con.commit()

	progress.set_progress(40, "Importing base packages manifests")
	importBaseManifests(installation)
	registerAtomicActions(installation)
	registerHost(installation)

def __registerComponent(con, installation, name, ctype):
	return uFakePackaging.registerComponent(con, 1, installation.importedBasePackages[(name, ctype) ], installation.rootpath)


class DBComponent:
	def __init__(self, con, component_id):
		self.con = con
		self.component_id = component_id

	def set_props(self, **kwds):
		for k in kwds:
			uPEM.setComponentProperty(self.component_id, k, kwds[k], connection = self.con)


def orb_endpoint(ip, port):
	return "iiop://%s:%s" % (ip, port)

def resolver_ior(ip, port):
	return 'corbaloc:iiop:%s:%s/ObjectResolver' % (ip, port)

ip_pattern = re.compile(r'(\d+)\.(\d+)\.(\d+)\.(\d+)')
def normalizeIP(ip):
	m = ip_pattern.match(ip)
	if not m:
		raise Exception("%s: is not valid ip address" % ip)
	return "%03d.%03d.%03d.%03d" % (int(m.group(1)), int(m.group(2)), int(m.group(3)), int(m.group(4)))


def netInterface(con, host_id, name):
	cur = con.cursor()
	cur.execute("SELECT if_id FROM net_interfaces WHERE if_name = %s AND host_id = %s", name, host_id)
	row = cur.fetchone()
	if row:
		return row[0]
	return con.insertRecordWithId("net_interfaces", host_id = host_id, if_name = name, if_description = name)


def getInterfaceProperties(con, component_id):
	cur = con.cursor()
	cur.execute("""SELECT i.service_type
		 FROM components c JOIN packages p ON (p.pkg_id = c.pkg_id) 
				 JOIN package_interfaces pi ON (pi.pkg_id = p.pkg_id)
				 JOIN interfaces i ON (i.interface_id = pi.interface_id)
				 WHERE c.component_id = %s""", component_id)

	interfaces = [ row[0] for row in cur.fetchall() ]

	return zip([ "interface%s" % num for num in xrange(len(interfaces))], interfaces)

pNamePath = [ 
	('conf_files_path', 'etc'), 
	('pid_files_path', '/var/run'), 
	('svc.conf', 'etc/svc.conf'), 
	('pleskd.props', 'etc/pleskd.props'), 
	('chdir', 'var/cores'), 
	('ipadd', 'bin/ipadd.sh'), 
	('ipdel', 'bin/ipdel.sh'), 
	('ipadd-fake', 'bin/ipadd-fake.sh'), 
	('ipdel-fake', 'bin/ipdel-fake.sh'), 
	('ifcfg-path', 'bin') ]

def registerHost(installation):
	progress = installation.progress

	progress.set_progress(60, "Registering management node")
	con = uSysDB.connect()
	cur = con.cursor()
	cur.execute("SELECT platform_id FROM platforms WHERE arch = %s AND opsys = %s AND osrel = %s", installation.platform.arch, installation.platform.os, installation.platform.osver)

	row = cur.fetchone()
	if not row:
		raise Exception("%s-%s-%s - unknown platform" % installation.platform)		
	platform_id = row[0]	
	
	host_id = con.insertRecordWithId("hosts", primary_name = installation.hostname, platform_id = platform_id, default_rootpath = installation.rootpath, state = 'm', proprietor_id = OID_ADMIN_ACCOUNT, platform_desc = "%s %s %s" % (installation.platform.arch, installation.platform.os, installation.platform.osver))
	con.insertRecord("management_nodes", host_id = host_id, in_sync = 'y', operational = 'y')

	con.insertRecord("hosts_extended_info", host_id = host_id)

	#generate certificate for HCL on MN
	subj = "/C=US/O=Parallels/CN=PEM_SlaveNode_" + str(host_id)
	b64privkey, b64cert = uPEM.generate_self_signed_certificate(subj)
	
	props  = installation.properties

	sysuid, sysgid = 0, 0
	if props.has_key('sysuserid') and props.has_key('sysgroupid'):
		sysuid, sysgid = props['sysuserid'], props['sysgroupid']

	ifname = props["communication_netiface_name"]
	netInterface(con, 1, ifname)

	ip = props['communication_ip']
	pem_config_cmp_id = __registerComponent(con, installation, 'pem_config', 'other')
	pem_config = DBComponent(con, pem_config_cmp_id)
	kernel_port = int( props['kernel_endpoint_port'])
	pem_config.set_props(kernel_resolver_ior = resolver_ior(ip, kernel_port),
		sysuser = props['sysuser'], sysgroup = props['sysgroup'],
		db_name = props['db'], db_host= props['db_hostname'],
		db_login = props['dsn_login'], db_passwd = props['dsn_passwd'],
		rdbms = { uSysDB.MSSQL : "MSSQL", uSysDB.PgSQL : "PostgreSQL"}[uSysDB.DBType] )

	__registerComponent(con, installation, 'libplatform', 'other')

	progress.set_progress(70, "Registering pleskd")
	pleskd_cmp_id = __registerComponent(con, installation, 'pleskd', 'other')

	pleskd_id = uFakePackaging.registerSC2(con = con, name = "pleskd", version = '1.1', user_id = OID_SYSTEM_ACCOUNT, sc_group_id = 0, loader_id = -1, location_id = "pleskd", component_id = pleskd_cmp_id)

	cur.execute("UPDATE hosts SET pleskd_id = %s, sn_certificate = '%s' WHERE host_id = %s" % (pleskd_id, b64cert, 1))

	for pname, path in pNamePath:
		uPEM.setComponentProperty(pleskd_cmp_id, pname, os.path.join(installation.rootpath, path), connection = con)

	pleskd = DBComponent(con, pleskd_cmp_id)

	if sys.platform == 'win32':
		pleskd.set_props(**{ 'windows.domain.name' : 'NT AUTHORITY',
			'windows.admin.user.login' : 'SYSTEM',
			'windows.admin.user.password' : '' ,
			'os.arch' : installation.platform.arch })

	#generate kernel public/private rsa keys
	b64der, b64derpub = uPEM.generate_rsa_key()
	pleskd.set_props(kernel_pub_key = b64derpub)
	
	#generate encryption key
	b64key = uPEM.generateEncryptionKey()
	pleskd.set_props(encryption_key = b64key)	
	
	#set RSA private key for HCL on MN
	key_dir = installation.rootpath + '/etc/keys'
	if not os.path.exists(key_dir):
		os.makedirs(key_dir)
		
	priv_key_path = key_dir + '/sn_private_key.pem'
	priv_key_file = open(priv_key_path, 'w+')
	priv_key_file.write(b64privkey)
	priv_key_file.close()

	pleskd.set_props(** {
			'communication.ip': ip,
			'orb.endpoint.port': props['pleskd_endpoint_port'],
			'orb.endpoint.addresses': ip})

	cur.execute("SELECT name, value FROM v_props WHERE component_id = %s", pleskd_cmp_id)
	pleskd_props = [ (row[0], row[1]) for row in cur.fetchall()]
	pleskd_props += [ ("sc_id", pleskd_id), ("host_id", 1), ('login', installation.system_login), ('passwd', installation.system_password) ]
	pleskd_props += getInterfaceProperties(con, pleskd_cmp_id)

	progress.set_progress(75, "Registering SoLoader")

	soloader_cmp_id = __registerComponent(con, installation, 'SoLoader', 'other')

	base_packages = findBasePackages(installation)
	chief_pkg = kernel_pkg = ppm_pkg = None
	for p in base_packages:
		if p.package.name == 'chief':
			chief_pkg = p
		elif p.package.name == 'Kernel':
			kernel_pkg = p
		elif p.package.name == 'ppm':
			ppm_pkg = p

	progress.set_progress(80, "Registering Kernel")

	kernel_cmp_id = __registerComponent(con, installation, 'Kernel', 'sc')
	uFakePackaging.registerSC(con, soloader_cmp_id, kernel_cmp_id, installation.importedBasePackages[('Kernel', 'sc')], installation.rootpath, kernel_pkg)
	kernel = DBComponent(con, kernel_cmp_id)
	kernel.set_props(dsn = props['dsn'], dsn_login = props['dsn_login'], dsn_passwd = props['dsn_passwd'], orb_endpoint = orb_endpoint(ip, kernel_port), kernel_resolver_ior = resolver_ior(ip, kernel_port))

	cur.execute("SELECT name, value FROM v_props WHERE component_id = %s", kernel_cmp_id)
	kernel_props = [ (row[0], row[1]) for row in cur.fetchall()]
	kernel_props += [("loader_id", soloader_cmp_id), ("location_id", uFakePackaging.location_id(kernel_pkg, installation.rootpath))]
	kernel_props += [ ("kernel_priv_key", b64der) ]
	kernel_props += getInterfaceProperties(con, kernel_cmp_id)
	if props.has_key("external_db_service"):
		kernel_props.append(("external_db_service", props["external_db_service"]))

	uUtil.createPropertiesFile(pleskd_props, os.path.join(installation.rootpath, 'etc', 'pleskd.props'), uid = sysuid, gid = sysgid)
	uUtil.createPropertiesFile(kernel_props, os.path.join(installation.rootpath, 'etc', 'Kernel.conf'), uid = sysuid, gid = sysgid)

	progress.set_progress(90, "Registering chief")

	chief_cmp_id = __registerComponent(con, installation, 'chief', 'sc')
	uFakePackaging.registerSC(con, soloader_cmp_id, chief_cmp_id, installation.importedBasePackages[('chief', 'sc') ],  installation.rootpath, chief_pkg)

	progress.set_progress(95, "Registering PPM")

	ppm_cmp_id = __registerComponent(con, installation, 'ppm', 'sc')
	uFakePackaging.registerSC(con, soloader_cmp_id, ppm_cmp_id, installation.importedBasePackages[('ppm', 'sc') ],  installation.rootpath, ppm_pkg)

	external_ip = normalizeIP(props['external_ip'])

	pool_id = con.insertRecordWithId("ip_pools", pool_name = "default pool", ip_from = external_ip, ip_to = external_ip, owner_id = OID_ADMIN_ACCOUNT, sharing = 'a', netmask=props["external_netmask"])
	external_iface_id = netInterface(con, 1, props['external_netiface_name'])

	mn_ip_id = con.insertRecordWithId("configured_ips",
		host_id = 1,
		account_id = OID_ADMIN_ACCOUNT,
		sub_id = 1,
		if_id = external_iface_id,
		ip_address = external_ip,
		netmask = props["external_netmask"],
		pool_id = pool_id,
		ip_rti_id = 0,
		traffic_rti_id = 0,
		state = "A",
		ip_type = "S"
		)

	con.insertRecord("used_ips", ip_id = mn_ip_id, obj_id = 1, obj_domain = "IPManagement", obj_kind = "Host")
	con.insertRecord("bound_ip_pools", if_id = external_iface_id, pool_id = pool_id)
	con.insertRecord("assigned_ip_pools", user_id = OID_ADMIN_USER, pool_id = pool_id)
	cur.execute("""INSERT INTO ip_pools_purposes(pool_id, purpose_id)
	SELECT %s, purpose_id
	FROM ip_pool_declared_purposes
	WHERE purpose IN ('COMMON.INTERNAL', 'SHARED.HOSTING');""", pool_id)


	cur.execute("""
	INSERT INTO observers (sc_id, interface_id)
	SELECT si.sc_id, i.interface_id
	FROM sc_interfaces si JOIN interfaces i ON (si.interface_id = i.interface_id)
	WHERE
	i.service_type LIKE '%Observer'
	AND NOT EXISTS (
	SELECT 1 FROM observers o WHERE
	o.sc_id = si.sc_id
	AND o.interface_id = i.interface_id);
	""")

	PEMVersion.recordInstalledVersions(installation.buildinfo.version_list)

	con.commit()
	progress.set_progress(100)

class ImportedPkgCounter(uUtil.CounterCallback):
	def __init__(self, iprogress, init_percents = 0, end_percents = 100):
		uUtil.CounterCallback.__init__(self)
		self.iprogress = iprogress
		self.init_percents = init_percents
		self.my_percents = end_percents - init_percents

	def new_item(self, name):
		uUtil.CounterCallback.new_item(self, name)
		if self.iprogress:
			self.iprogress.set_progress(self.init_percents + self.my_percents * self.processed / self.total, "%s" % name)


class SCWaiter:
	def __init__(self, to_wait, progress, init_percents = 0, end_percents = 100):
		self.progress = progress
		self.init_percents = init_percents
		self.my_percents = end_percents - init_percents
		self.to_wait = to_wait
		self.total = len(to_wait)
		self.show_progress()

	def show_progress(self):
		self.progress.set_progress(self.init_percents + self.my_percents * (self.total - len(self.to_wait)) / self.total, "Waiting for %s to start" % ' '.join(self.to_wait))

	def started(self, scs):
		for sc in scs:
			if sc in self.to_wait:
				self.to_wait.remove(sc)
		self.show_progress()

def importPackages(installation):
	installation.progress.new_phase("Importing packages")
	installation.progress.set_progress(0, "Starting the system")
	if sys.platform != 'win32':
		uPEM.startMN()

	installation.latestPackages = uPackaging.getNewestPackages([installation.build])
	uPEM.wait_for_scs(["chief", "ppm"], monitor = SCWaiter(["chief", "ppm"], installation.progress, 0, 2))

	uPackaging.importPackages(installation.latestPackages, installation.rootpath, installation.build.topdir, ImportedPkgCounter(installation.progress, 2, 99))

	props = installation.properties
	if sys.platform == 'win32':
		mirror_url = 'http://%(ipemts_user)s:%(ipemts_passwd)s@%(communication_ip)s:%(ipemts_port)s/tarballs' % props
	else:
		mirror_url = 'http://%(communication_ip)s/tarballs' % props

	# create PPM mirror on MN
	localpath = os.path.join(props["rootpath"], "install", "tarballs")
	uPEM.execCtl("ppm_ctl", "mirror", "create", "1", localpath, mirror_url)

	installation.progress.set_progress(100)

def copyAPS(props):
	rootpath = props['rootpath']
	tdir = os.path.join(rootpath, "install", "APS")
	if not tdir.endswith('/'):
		tdir += '/'
	try:
		os.makedirs(tdir)
	except OSError, e:
		if e.errno != errno.EEXIST:
			raise
	source_dir = props.get('source_dir', None)

	if source_dir is not None:
		os.path.walk(source_dir, copyApplications, tdir)

aps_pattern = re.compile(r'.+\.app\.zip$')
def copyApplications(to, dirname, fnames):
	for fn in fnames:
		if aps_pattern.match(fn):
			uUtil.ln_cp(os.path.join(dirname, fn), to)

package_propname_pattern = re.compile(r"(sc|cp|other)-(.+):(.+)")


def installModule(installation, module):
	installation.progress.new_phase("Installing module %s" % module.description)
	pkg_properties = {}
	for p in installation.properties:
		m = package_propname_pattern.match(p)
		if m is not None:
			pkg = m.group(1), m.group(2)
			propname = m.group(3)
			pvalue = installation.properties[p]
			if pkg not in pkg_properties:
				pkg_properties[pkg] = {}

			pkg_properties[pkg][propname] = pvalue

	install_ui = installation.properties.get("install_ui", "yes") == "yes"

	if install_ui:
		install_where = (uDLModel.MN_pkg, uDLModel.UI_pkg )
	else:
		install_where = (uDLModel.MN_pkg, )

	if module.name == 'Platform':
		# install Platform SC's and config packages sequentially as usual (they've got some implicit dependencies, that ppm is not able to resolve)

		to_install = [pkg for pkg in module.packages if pkg.where_pkg in install_where and (pkg.ctype == 'sc' or pkg.name in ['os_config', 'rc_config'])]

		reposReady = False
		total = len(to_install)
		go_installed = open_api_installed = module.name != 'Platform'
		for i in xrange(0, total):
			package = to_install[i]
			ctype, name = package.ctype, package.name
			installation.progress.set_progress(30 * i / total, "%s-%s" % (ctype, name))
			try:
				if go_installed and open_api_installed:
					uPackaging.installPackageToHostAPI(1,
						name = name, ctype = ctype,
						properties = pkg_properties.get((ctype, name), {}))
				else:
					uPackaging.installPackageToHost(1, name, ctype, pkg_properties.get((ctype, name), {}))
					if name == 'group_operations':
						go_installed = True
					elif name == 'OpenAPI':
						open_api_installed = True

					if go_installed and open_api_installed and reposReady == False:
						# create native repos as soon as OpenAPI is available, so that platform components could resolve their native dependencies
						# TODO move OpenAPI to core and make core deployment more obvious
						uPackaging.createAndUpdateNativeRepos([installation.build], installation.progress)
						reposReady = True

				if "sc" == ctype:
					# if this was SC, that we just installed, wait for it to be fully functional before installation of another one
					uPEM.wait_for_scs([name], monitor = SCWaiter([name], installation.progress, 0, 2))

			except uPackaging.PackageNotFound:
				uLogging.err("%s-%s - cannot find package", ctype, name)
	else:
		# install other SCs using meta-package		
		installation.progress.set_progress(0, "%s" % ('Service controllers'))

		to_install = [(pkg.name, pkg.ctype) for pkg in module.packages if pkg.where_pkg in install_where and pkg.ctype == 'sc']
		uFakePackaging.installMetaPackage(1, to_install)

	installation.progress.set_progress(30, "%s" % ('Privileges'))

	to_install = [(pkg.name, pkg.ctype) for pkg in module.packages if pkg.where_pkg in install_where and 'Privilege' in pkg.name]
	uFakePackaging.installMetaPackage(1, to_install)

	installation.progress.set_progress(60, "%s" % ('Control panels and other'))

	to_install = [(pkg.name, pkg.ctype) for pkg in module.packages if pkg.where_pkg in install_where and pkg.ctype != 'sc']
	uFakePackaging.installMetaPackage(1, to_install)

	installation.progress.set_progress(100, "%s" % ('Done'))

def registerModule(topdir, module, installed):
	api = openapi.OpenAPI()
	f = open(os.path.join(topdir, module.source), "r")
	xml = f.read()
	f.close()
	api.pem.packaging.registerModule(mdl = xml, installed = installed)

def installPlatform(installation):
	platform = [ m for m in installation.build.modules if m.name == 'Platform']
	if not platform:
		raise Exception("Cannot find Platform module")
	platform = platform[0]
	installModule(installation, platform)
	registerModule(installation.build.topdir, platform, True)

	progress = installation.progress
	progress.set_progress(100, "Applying system users")
	completeFirstUsersCreation(installation)

def installModules(installation):
	mprop = installation.properties.get("modules", None)
	uLogging.notice("modules = %s", mprop)

	if mprop:
		if mprop.lower() == 'all':
			module_names = [ m.name for m in installation.build.modules if m.name != 'Platform' ]
		else:
			module_names = [ m.strip() for m in mprop.split(",")]
			# filter out Platform, no need to install second time
			module_names = [ m for m in module_names if m != 'Platform' ]
	else:
		module_names = []

	existing_modules = {}
	for module in installation.build.modules:
		existing_modules[module.name] = module

	uLogging.notice("existing: %s, module_names: %s", existing_modules, module_names)

	for name in module_names:
		if name not in existing_modules:
			uLogging.err("Cannot find module %s in build", name)
		else:
			module = existing_modules[name]
			installModule(installation, module)
			registerModule(installation.build.topdir, module, True)

	for n in existing_modules:
		if n not in module_names and n != 'Platform':
			registerModule(installation.build.topdir, existing_modules[n], False)

	install_ui = installation.properties.get("install_ui", "yes") == "yes"
	if install_ui:
		uPEM.stopUI(1)
		uPEM.startUI(1)

def finalize(installation):
	dummy = installation
